/*
 * Copyright (c) 2002-2005 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  o Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *     
 *  o Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution. 
 *     
 *  o Neither the name of JGoodies Karsten Lentzsch nor the names of 
 *    its contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission. 
 *     
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

package com.jgoodies.binding.extras;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import com.jgoodies.binding.value.ValueHolder;
import com.jgoodies.binding.value.ValueModel;

/**
 * This implementation of the <code>TableModel</code> interface 
 * delegates all requests to an underlying TableModel delegate.
 * It can be used to change a <code>JTable</code>'s model while the
 * table is beeing displayed.<p>
 * 
 * <strong>Note:</strong> This class is not yet part of the binary Binding 
 * library; it comes with the Binding distributions as an extra. 
 * <strong>The API is work in progress and may change without notice; 
 * this class may even be completely removed from future distributions.</strong> 
 * If you want to use this class, you may consider copying it into 
 * your codebase.
 * 
 * @author Karsten Lentzsch
 * @version $Revision: 1.6 $
 */
public final class DelegatingTableModel extends AbstractTableModel {

    /**
     * Holds a model that in turn holds the delegated <code>TableModel</code>.
     */
    private final ValueModel delegateChannel;


    // Instance Creation ******************************************************
    
    /**
     * Constructs a <code>TableAdaptor</code> that holds the delegated
     * <code>TableModel</code> in the given delegate channel.
     * 
     * @param delegateChannel    a <code>ValueModel</code> that holds the
     *     <code>TableModel</code>
     * @throws NullPointerException   if delegateChannel is null
     */
    public DelegatingTableModel(ValueModel delegateChannel) {
        this.delegateChannel = delegateChannel;
        if (delegateChannel == null) 
            throw new NullPointerException("The delegate channel must not be null.");
             
        delegateChannel.addValueChangeListener(new DelegateChangeHandler());
    }
    
    /**
     * Constructs a <code>TableAdaptor</code> on the given 
     * delegate <code>TableModel</code>.
     * 
     * @param delegate    a <code>TableModel</code> that is used as initial
     *     <code>TableModel</code> delegate
     */
    public DelegatingTableModel(TableModel delegate) {
        this(new ValueHolder(delegate));
    }

    /**
     * Constructs a <code>TableAdaptor</code> on a dummy <code>TableModel</code>.
     */
    public DelegatingTableModel() {
        this(new DummyTableModel());
    }
    
    
    // Accessing the Delegate ************************************************

    /**
     * Returns the current <code>TableModel</code> delegate.
     * 
     * @return the current <code>TableModel</code> delegate
     */
    public TableModel getDelegate() {
        return (TableModel) delegateChannel.getValue();
    }
    
    /**
     * Sets a new <code>TableModel</code> delegate.
     * 
     * @param newDelegate    the new <code>TableModel</code> delegate
     */
    public void setDelegate(TableModel newDelegate) {
        delegateChannel.setValue(newDelegate);
    }


    // TableModel Implementation **********************************************
    
    /**
     * Returns the most specific superclass for all the cell values 
     * in the column.  This is used by the <code>JTable</code> to set up a 
     * default renderer and editor for the column.
     *
     * @param columnIndex  the index of the column
     * @return the common ancestor class of the object values in the model.
     */
    public Class getColumnClass(int columnIndex) {
        return getDelegate().getColumnClass(columnIndex);
    }

    /**
     * Returns the number of columns in the model. A
     * <code>JTable</code> uses this method to determine how many columns it
     * should create and display by default.
     *
     * @return the number of columns in the model
     * @see #getRowCount()
     */
    public int getColumnCount() {
        return getDelegate().getColumnCount();
    }

    /**
     * Returns the name of the column at <code>columnIndex</code>.  This is used
     * to initialize the table's column header name.  Note: this name does
     * not need to be unique; two columns in a table can have the same name.
     *
     * @param   columnIndex the index of the column
     * @return  the name of the column
     */
    public String getColumnName(int columnIndex) {
        return getDelegate().getColumnName(columnIndex);
    }

    /**
     * Returns the number of rows in the model. A
     * <code>JTable</code> uses this method to determine how many rows it
     * should display.  This method should be quick, as it
     * is called frequently during rendering.
     *
     * @return the number of rows in the model
     * @see #getColumnCount()
     */
    public int getRowCount() {
        return getDelegate().getRowCount();
    }

    /**
     * Returns the value for the cell at <code>columnIndex</code> and
     * <code>rowIndex</code>.
     *
     * @param   rowIndex    the row whose value is to be queried
     * @param   columnIndex     the column whose value is to be queried
     * @return  the value Object at the specified cell
     */
    public Object getValueAt(int rowIndex, int columnIndex) {
        return getDelegate().getValueAt(rowIndex, columnIndex);
    }
    
    
    // Helper Classes ********************************************************

    /**
     * Listens to changes of the subject.
     */
    private class DelegateChangeHandler implements PropertyChangeListener {

        /**
         * The delegate has been changed. Fire a table data change.
         * 
         * @param evt   the property change event to be handled
         */
        public void propertyChange(PropertyChangeEvent evt) {
            fireTableDataChanged();
        }
    }


    /**
     * A dummy model that can be used by the default constructor.
     */
    private static class DummyTableModel extends AbstractTableModel {
        
        /**
         * Returns a dummy column name for every column index.
         * 
         * @param columnIndex   the index of the requested column
         * @return a dummy string
         */
        public String getColumnName(int columnIndex) {
            return "Dummy";
        }
        
        /**
         * Returns a dummy row count zero.
         * 
         * @return 0
         */
        public int getRowCount() {
            return 0;
        }
        
        /**
         * Returns a dummy column count of 1.
         * 
         * @return 1
         */
        public int getColumnCount() {
            return 1;
        }
        
        /**
         * Returns the empty string as dummy value for all cells.
         * 
         * @param rowIndex   the row's index
         * @param columnIndex   the column's index
         * @return an empty string as dummy value
         */
        public Object getValueAt(int rowIndex, int columnIndex) {
            return "";
        }
    }

}
