/*
 * Copyright (c) 2002-2005 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  o Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *     
 *  o Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution. 
 *     
 *  o Neither the name of JGoodies Karsten Lentzsch nor the names of 
 *    its contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission. 
 *     
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

package com.jgoodies.binding.extras;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import javax.swing.ListModel;
import javax.swing.event.EventListenerList;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

import com.jgoodies.binding.beans.ExtendedPropertyChangeSupport;
import com.jgoodies.binding.value.AbstractValueModel;
import com.jgoodies.binding.value.ValueModel;

/**
 * A {@link com.jgoodies.binding.value.ValueHolder} implementation that holds 
 * instances of {@link javax.swing.ListModel} and that is a 
 * <code>ListModel</code> itself.<p>
 * 
 * <strong>Note:</strong> This class is not yet part of the binary Binding 
 * library; it comes with the Binding distributions as an extra. 
 * <strong>The API is work in progress and may change without notice; 
 * this class may even be completely removed from future distributions.</strong> 
 * If you want to use this class, you may consider copying it into 
 * your codebase.
 * 
 * @author Karsten Lentzsch
 * @author Mattias Neuling
 * @version $Revision: 1.10 $
 */
public final class ListModelHolder implements ValueModel, ListModel {

    /** 
     * The name of the bound property <em>value</em>.    
     */
    private static final String PROPERTYNAME_VALUE = AbstractValueModel.PROPERTYNAME_VALUE;

    /**
     * Holds the change handler that fires changes is the underlying
     * list model changes.
     */
    private final ListDataChangeHandler listDataChangeHandler;

    /**
     * If any <code>PropertyChangeListeners</code> have been registered,
     * the <code>changeSupport</code> field describes them.
     *
     * @see #addPropertyChangeListener(String, PropertyChangeListener)
     * @see #removePropertyChangeListener(String, PropertyChangeListener)
     */
    private ExtendedPropertyChangeSupport changeSupport;

    /**
     * Holds the list model value.
     */
    private ListModel listModel;


    // Instance Creation ******************************************************

    /**
     * Constructs a ListModelHolder on the given ListModel.
     * 
     * @param listModel   the initial list model
     */
    public ListModelHolder(ListModel listModel) {
        listDataChangeHandler = new ListDataChangeHandler();
        setValue(listModel);
    }


    // ValueModel Implementation **********************************************

    /**
     * Returns the observed value. In case of a write-only value,
     * implementors may choose to either reject this operation or 
     * to do nothing.
     * 
     * @return the observed value
     */
    public Object getValue() {
        return listModel;
    }


    /**
     * Sets a new value and notifies any registered value listeners 
     * if the value has changed. 
     * 
     * @param newValue  the ListModel to be set
     * 
     * @throws ClassCastException if the new value is not a <code>ListModel</code>
     */
    public void setValue(Object newValue) {
        setListModel((ListModel) newValue);
    }


    /**
     * Sets a new list model  and notifies any registered value listeners 
     * if it has changed. 
     * 
     * @param newListModel  the <code>ListModel</code> to be set
     */
    public void setListModel(ListModel newListModel) {
        ListModel oldListModel = listModel;
        if (oldListModel != null) {
            oldListModel.removeListDataListener(listDataChangeHandler);
        }
        listModel = newListModel;
        if (newListModel != null) {
            newListModel.addListDataListener(listDataChangeHandler);
        }
        firePropertyChange(PROPERTYNAME_VALUE, oldListModel, newListModel);
        fireContentsChanged(-1, -1);
    }


    /**
     * Adds the given PropertyChangeListener that 
     * will be notified if the value has been changed.<p>
     * 
     * Changes in the value are checked using <code>==</code> 
     * not <code>#equals</code>. This ensures that this class' internal 
     * <code>ListDataListener</code> can be moved properly whenever
     * the <code>ListModel</code> instance changes. 
     * 
     * @param listener  the listener to be added
     * 
     * @see ExtendedPropertyChangeSupport
     */
    public synchronized void addValueChangeListener(
            PropertyChangeListener listener) {
        addPropertyChangeListener(PROPERTYNAME_VALUE, listener);
    }


    /**
     * Removes the given PropertyChangeListener.
     * 
     * @param listener  the listener to be removed
     */
    public synchronized void removeValueChangeListener(
            PropertyChangeListener listener) {
        removePropertyChangeListener(PROPERTYNAME_VALUE, listener);
    }


    // ValueHolder Helper Code ************************************************

    /**
     * Adds a PropertyChangeListener to the listener list for a specific
     * property. The specified property may be user-defined.<p>
     * 
     * Note that if this Model is inheriting a bound property, then no event
     * will be fired in response to a change in the inherited property.<p>
     * 
     * If listener is <code>null</code>, no exception is thrown 
     * and no action is performed.<p>
     * 
     * The <code>ExtendedPropertyChangeSupport</code> instance created here
     * is configured to check value changes using <code>==</code>, 
     * not <code>#equals</code>. This ensures that this class' internal 
     * <code>ListDataListener</code> can be moved properly whenever
     * the <code>ListModel</code> instance changes. 
     *
     * @param propertyName      one of the property names listed above
     * @param listener          the PropertyChangeListener to be added
     *
     * @see #removePropertyChangeListener(String, PropertyChangeListener)
     * @see #addPropertyChangeListener(String, PropertyChangeListener)
     * @see ExtendedPropertyChangeSupport
     */
    private synchronized void addPropertyChangeListener(
            String propertyName, PropertyChangeListener listener) {
        if (listener == null) { return; }
        if (changeSupport == null) {
            changeSupport = new ExtendedPropertyChangeSupport(this, true);
        }
        changeSupport.addPropertyChangeListener(propertyName, listener);
    }


    /**
     * Removes a PropertyChangeListener from the listener list for a specific
     * property. This method should be used to remove PropertyChangeListeners
     * that were registered for a specific bound property.<p>
     * 
     * If listener is <code>null</code>, no exception is thrown 
     * and no action is performed.
     *
     * @param propertyName      a valid property name
     * @param listener          the PropertyChangeListener to be removed
     *
     * @see #addPropertyChangeListener(String, PropertyChangeListener)
     */
    private synchronized void removePropertyChangeListener(
            String propertyName, PropertyChangeListener listener) {
        if (listener == null || changeSupport == null) { return; }
        changeSupport.removePropertyChangeListener(propertyName, listener);
    }


    /**
     * Support for reporting bound property changes for Object properties. 
     * This method can be called when a bound property has changed and it will
     * send the appropriate PropertyChangeEvent to any registered
     * PropertyChangeListeners.
     *
     * @param propertyName      the property whose value has changed
     * @param oldValue          the property's previous value
     * @param newValue          the property's new value
     */
    private void firePropertyChange(String propertyName, Object oldValue,
            Object newValue) {
        PropertyChangeSupport aChangeSupport = changeSupport;
        if (aChangeSupport == null) { return; }
        aChangeSupport.firePropertyChange(propertyName, oldValue, newValue);
    }

    // ListModel Field ********************************************************

    /**
     * Holds the ListDataListeners registered with this ListModel implementation.
     * 
     * @see #addListDataListener(ListDataListener)
     * @see #removeListDataListener(ListDataListener)
     */
    private EventListenerList listenerList = new EventListenerList();


    // ListModel Implementation ***********************************************

    /**
     * Adds a listener to the list that's notified each time a change
     * to the data model occurs.
     *
     * @param l the <code>ListDataListener</code> to be added
     */
    public void addListDataListener(ListDataListener l) {
        getEventListenerList().add(ListDataListener.class, l);
    }


    /**
     * Removes a listener from the list that's notified each time a 
     * change to the data model occurs.
     *
     * @param l the <code>ListDataListener</code> to be removed
     */
    public void removeListDataListener(ListDataListener l) {
        getEventListenerList().remove(ListDataListener.class, l);
    }


    /**
     * Returns the value at the specified index, or <code>null</code>
     * if the current value is <code>null</code>.
     * 
     * @param index the requested index
     * @return the value at <code>index</code> or <code>null</code>
     *     if the current value is <code>null</code> 
     */
    public Object getElementAt(int index) {
        return listModel == null ? null : listModel.getElementAt(index);
    }


    /** 
     * Returns the length of the list or 0 if there's no list.
     * 
     * @return the length of the list or 0 if there's no list
     */
    public int getSize() {
        return listModel == null ? 0 : listModel.getSize();
    }


    /**
     * Returns an array of all the list data listeners
     * registered on this <code>ArrayListModel</code>.
     *
     * @return all of this model's <code>ListDataListener</code>s,
     *         or an empty array if no list data listeners
     *         are currently registered
     * 
     * @see #addListDataListener(ListDataListener)
     * @see #removeListDataListener(ListDataListener)
     */
    public ListDataListener[] getListDataListeners() {
        return (ListDataListener[]) getEventListenerList()
                .getListeners(ListDataListener.class);
    }


    // ListModel Helper Code **************************************************

    /**
     * This method must be called <em>after</em> one or more elements 
     * of the list change.  The changed elements
     * are specified by the closed interval index0, index1 -- the end points
     * are included.  Note that index0 need not be less than or equal to index1.
     * 
     * @param index0 one end of the new interval
     * @param index1 the other end of the new interval
     * @see EventListenerList
     */
    private void fireContentsChanged(int index0, int index1) {
        Object[] listeners = getEventListenerList().getListenerList();
        ListDataEvent e = null;

        for (int i = listeners.length - 2; i >= 0; i -= 2) {
            if (listeners[i] == ListDataListener.class) {
                if (e == null) {
                    e = new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED,
                            index0, index1);
                }
                ((ListDataListener) listeners[i + 1]).contentsChanged(e);
            }
        }
    }


    /**
     * This method must be called <em>after</em> one or more elements 
     * are added to the model.  The new elements
     * are specified by a closed interval index0, index1 -- the end points
     * are included.  Note that index0 need not be less than or equal to index1.
     * 
     * @param index0 one end of the new interval
     * @param index1 the other end of the new interval
     * @see EventListenerList
     */
    private void fireIntervalAdded(int index0, int index1) {
        Object[] listeners = getEventListenerList().getListenerList();
        ListDataEvent e = null;

        for (int i = listeners.length - 2; i >= 0; i -= 2) {
            if (listeners[i] == ListDataListener.class) {
                if (e == null) {
                    e = new ListDataEvent(this, ListDataEvent.INTERVAL_ADDED,
                            index0, index1);
                }
                ((ListDataListener) listeners[i + 1]).intervalAdded(e);
            }
        }
    }


    /**
     * This method must be called <em>after</em>  one or more elements 
     * are removed from the model. 
     * <code>index0</code> and <code>index1</code> are the end points
     * of the interval that's been removed.  Note that <code>index0</code>
     * need not be less than or equal to <code>index1</code>.
     * 
     * @param index0 one end of the removed interval,
     *               including <code>index0</code>
     * @param index1 the other end of the removed interval,
     *               including <code>index1</code>
     * @see EventListenerList
     */
    private void fireIntervalRemoved(int index0, int index1) {
        Object[] listeners = getEventListenerList().getListenerList();
        ListDataEvent e = null;

        for (int i = listeners.length - 2; i >= 0; i -= 2) {
            if (listeners[i] == ListDataListener.class) {
                if (e == null) {
                    e = new ListDataEvent(this, ListDataEvent.INTERVAL_REMOVED,
                            index0, index1);
                }
                ((ListDataListener) listeners[i + 1]).intervalRemoved(e);
            }
        }
    }

    
    /**
     * Lazily initializes and returns the event listener list used
     * to notify registered listeners.
     * 
     * @return the event listener list used to notify listeners
     */
    private EventListenerList getEventListenerList() {
        if (listenerList == null) {
            listenerList = new EventListenerList();
        }
        return listenerList;
    }
    
    
    // Helper Class ***********************************************************

    /**
     * Listens to subject changes and fires a contents change event.
     */
    private class ListDataChangeHandler implements ListDataListener {

        /** 
         * Sent after the indices in the index0, index1 
         * interval have been inserted in the data model.
         * The new interval includes both index0 and index1.
         *
         * @param evt  a <code>ListDataEvent</code> encapsulating the
         *    event information
         */
        public void intervalAdded(ListDataEvent evt) {
            fireIntervalAdded(evt.getIndex0(), evt.getIndex1());
        }


        /**
         * Sent after the indices in the index0, index1 interval
         * have been removed from the data model.  The interval 
         * includes both index0 and index1.
         *
         * @param evt  a <code>ListDataEvent</code> encapsulating the
         *    event information
         */
        public void intervalRemoved(ListDataEvent evt) {
            fireIntervalRemoved(evt.getIndex0(), evt.getIndex1());
        }


        /** 
         * Sent when the contents of the list has changed in a way 
         * that's too complex to characterize with the previous 
         * methods. For example, this is sent when an item has been
         * replaced. Index0 and index1 bracket the change.
         *
         * @param evt  a <code>ListDataEvent</code> encapsulating the
         *    event information
         */
        public void contentsChanged(ListDataEvent evt) {
            fireContentsChanged(evt.getIndex0(), evt.getIndex1());
        }

    }

}
